@@ -1,7 +1,9 @@
 from __future__ import annotations
 
 import json
+import tempfile
 from enum import Enum
+from pathlib import Path
 from typing import Any
 from typing import Dict
 from typing import Generator
@@ -12,19 +14,25 @@ from typing import Tuple
 from uuid import UUID
 
 import httpx
+import typer
 
 import kleinkram.errors
 from kleinkram._version import __version__
-from kleinkram.api.client import CLI_VERSION_HEADER
 from kleinkram.api.client import AuthenticatedClient
-from kleinkram.api.deser import FileObject
+from kleinkram.api.client import CLI_VERSION_HEADER
+from kleinkram.api.deser import (
+    FileObject,
+    _parse_run,
+    RunObject,
+    _parse_action_template,
+)
 from kleinkram.api.deser import MissionObject
 from kleinkram.api.deser import ProjectObject
 from kleinkram.api.deser import _parse_file
 from kleinkram.api.deser import _parse_mission
 from kleinkram.api.deser import _parse_project
 from kleinkram.api.pagination import paginated_request
-from kleinkram.api.query import FileQuery
+from kleinkram.api.query import FileQuery, RunQuery
 from kleinkram.api.query import MissionQuery
 from kleinkram.api.query import ProjectQuery
 from kleinkram.api.query import file_query_is_unique
@@ -38,12 +46,15 @@ from kleinkram.errors import InvalidMissionQuery
 from kleinkram.errors import InvalidProjectQuery
 from kleinkram.errors import MissionExists
 from kleinkram.errors import MissionNotFound
+from kleinkram.errors import MissionValidationError
 from kleinkram.errors import ProjectExists
 from kleinkram.errors import ProjectNotFound
-from kleinkram.models import File
+from kleinkram.errors import ProjectValidationError
+from kleinkram.models import File, Run, ActionTemplate
 from kleinkram.models import Mission
 from kleinkram.models import Project
 from kleinkram.utils import is_valid_uuid4
+from kleinkram.utils import split_args
 
 __all__ = [
     "_get_api_version",
@@ -79,6 +90,8 @@ PROJECT_ENDPOINT = "/projects"
 
 TAG_TYPE_BY_NAME = "/tag/filtered"
 
+ACTION_ENDPOINT = "/action"
+
 
 class Params(str, Enum):
     FILE_PATTERNS = "filePatterns"
@@ -159,15 +172,52 @@ def get_projects(
     client: AuthenticatedClient,
     project_query: ProjectQuery,
     max_entries: Optional[int] = None,
+    exact_match: bool = False,
 ) -> Generator[Project, None, None]:
     params = _project_query_to_params(project_query)
     response_stream = paginated_request(
-        client, PROJECT_ENDPOINT, params=params, max_entries=max_entries
+        client,
+        PROJECT_ENDPOINT,
+        params=params,
+        max_entries=max_entries,
+        exact_match=exact_match,
     )
     yield from map(lambda p: _parse_project(ProjectObject(p)), response_stream)
 
 
-def get_project(client: AuthenticatedClient, query: ProjectQuery) -> Project:
+LIST_ACTIONS_ENDPOINT = "/action/listActions"
+
+
+def get_runs(
+    client: AuthenticatedClient,
+    query: RunQuery,
+) -> Generator[Run, None, None]:
+
+    response_stream = paginated_request(client, LIST_ACTIONS_ENDPOINT)
+    yield from map(lambda p: _parse_run(RunObject(p)), response_stream)
+
+
+def get_run(
+    client: AuthenticatedClient,
+    run_id: str,
+) -> Run:
+    resp = client.get(f"{ACTION_ENDPOINT}/details", params={"uuid": run_id})
+    if resp.status_code == 404:
+        raise kleinkram.errors.RunNotFound(f"Run not found: {run_id}")
+    resp.raise_for_status()
+    return _parse_run(RunObject(resp.json()))
+
+
+def get_action_templates(
+    client: AuthenticatedClient,
+) -> Generator[ActionTemplate, None, None]:
+    response_stream = paginated_request(client, "/action/listTemplates")
+    yield from map(lambda p: _parse_action_template(RunObject(p)), response_stream)
+
+
+def get_project(
+    client: AuthenticatedClient, query: ProjectQuery, exact_match: bool = False
+) -> Project:
     """\
     get a unique project by specifying a project spec
     """
@@ -176,11 +226,39 @@ def get_project(client: AuthenticatedClient, query: ProjectQuery) -> Project:
             f"Project query does not uniquely determine project: {query}"
         )
     try:
-        return next(get_projects(client, query))
+        return next(get_projects(client, query, exact_match=exact_match))
     except StopIteration:
         raise ProjectNotFound(f"Project not found: {query}")
 
 
+def submit_action(
+    client: AuthenticatedClient, mission_uuid: UUID, template_uuid: UUID
+) -> str:
+    """
+    Submits a new action to the API and returns the action UUID.
+
+    Raises:
+        httpx.HTTPStatusError: If the API returns an error.
+        KeyError: If the response is missing 'actionUUID'.
+    """
+    submit_payload = {
+        "missionUUID": str(mission_uuid),
+        "templateUUID": str(template_uuid),
+    }
+
+    typer.echo("Submitting action...")
+    resp = client.post(f"{ACTION_ENDPOINT}/submit", json=submit_payload)
+    resp.raise_for_status()  # Raises on 4xx/5xx responses
+
+    response_data = resp.json()
+    action_uuid_str = response_data.get("actionUUID")
+
+    if not action_uuid_str:
+        raise KeyError("API response missing 'actionUUID'")
+
+    return action_uuid_str
+
+
 def get_mission(client: AuthenticatedClient, query: MissionQuery) -> Mission:
     """\
     get a unique mission by specifying a mission query
@@ -220,32 +298,9 @@ def _mission_name_is_available(
     return False
 
 
-def _project_name_is_available(client: AuthenticatedClient, project_name: str) -> bool:
-    project_query = ProjectQuery(patterns=[project_name])
-    try:
-        _ = get_project(client, project_query)
-    except ProjectNotFound:
-        return True
-    return False
-
-
-def _create_mission(
-    client: AuthenticatedClient,
-    project_id: UUID,
-    mission_name: str,
-    *,
-    metadata: Optional[Dict[str, str]] = None,
-    ignore_missing_tags: bool = False,
-) -> UUID:
-    """\
-    creates a new mission with the given name and project_id
-
-    if check_exists is True, the function will return the existing mission_id,
-    otherwise if the mission already exists an error will be raised
-    """
-    if metadata is None:
-        metadata = {}
-
+def _validate_mission_name(
+    client: AuthenticatedClient, project_id: UUID, mission_name: str
+) -> None:
     if not _mission_name_is_available(client, mission_name, project_id):
         raise MissionExists(
             f"Mission with name: `{mission_name}` already exists"
@@ -258,6 +313,91 @@ def _create_mission(
             "mission names must not be valid UUIDv4's"
         )
 
+    if mission_name.endswith(" "):
+        raise ValueError(
+            "A mission name cannot end with a whitespace. "
+            f"The given mission name was '{mission_name}'"
+        )
+
+
+def _project_name_is_available(client: AuthenticatedClient, project_name: str) -> bool:
+    project_query = ProjectQuery(patterns=[project_name])
+    try:
+        _ = get_project(client, project_query, exact_match=True)
+    except ProjectNotFound:
+        return True
+    return False
+
+
+def _validate_mission_created(
+    client: AuthenticatedClient, project_id: str, mission_name: str
+) -> None:
+    """
+    validate that a mission is successfully created
+    """
+    mission_ids, mission_patterns = split_args([mission_name])
+    project_ids, project_patterns = split_args([project_id])
+
+    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
+    mission_query = MissionQuery(
+        ids=mission_ids,
+        patterns=mission_patterns,
+        project_query=project_query,
+    )
+    try:
+        with tempfile.NamedTemporaryFile(suffix=".mcap", delete=False) as tmp:
+            tmp.write(b"dummy content")
+            tmp_path = Path(tmp.name)
+
+        kleinkram.core.upload(
+            client=client,
+            query=mission_query,
+            file_paths=[tmp_path],
+            verbose=False,
+        )
+
+        file_query = FileQuery(
+            ids=[],
+            patterns=[tmp_path.name],
+            mission_query=mission_query,
+        )
+        file_parsed = get_file(client, file_query)
+
+        kleinkram.core.delete_files(client=client, file_ids=[file_parsed.id])
+
+    except Exception as e:
+        raise MissionValidationError(f"Mission validation failed: {e}")
+
+    finally:
+        if tmp_path.exists():
+            tmp_path.unlink()
+
+
+def _create_mission(
+    client: AuthenticatedClient,
+    project_id: UUID,
+    mission_name: str,
+    *,
+    metadata: Optional[Dict[str, str]] = None,
+    ignore_missing_tags: bool = False,
+    required_tags: Optional[List[str]] = None,
+) -> UUID:
+    """\
+    creates a new mission with the given name and project_id
+
+    if check_exists is True, the function will return the existing mission_id,
+    otherwise if the mission already exists an error will be raised
+    """
+    if metadata is None:
+        metadata = {}
+
+    _validate_mission_name(client, project_id, mission_name)
+
+    if required_tags and not set(required_tags).issubset(metadata.keys()):
+        raise InvalidMissionMetadata(
+            f"Mission tags `{required_tags}` are required but missing from metadata: {metadata}"
+        )
+
     # we need to translate tag keys to tag type ids
     tags = _get_tags_map(client, metadata)
 
@@ -267,9 +407,9 @@ def _create_mission(
         "tags": {str(k): v for k, v in tags.items()},
         "ignoreTags": ignore_missing_tags,
     }
-
     resp = client.post(CREATE_MISSION, json=payload)
     resp.raise_for_status()
+    _validate_mission_created(client, str(project_id), mission_name)
 
     return UUID(resp.json()["uuid"], version=4)
 
@@ -277,10 +417,8 @@ def _create_mission(
 def _create_project(
     client: AuthenticatedClient, project_name: str, description: str
 ) -> UUID:
-    if not _project_name_is_available(client, project_name):
-        raise ProjectExists(f"Project with name: `{project_name}` already exists")
 
-    # TODO: check name and description are valid
+    _validate_project_name(client, project_name, description)
     payload = {"name": project_name, "description": description}
     resp = client.post(CREATE_PROJECT, json=payload)
     resp.raise_for_status()
@@ -288,18 +426,52 @@ def _create_project(
     return UUID(resp.json()["uuid"], version=4)
 
 
+def _validate_project_name(
+    client: AuthenticatedClient, project_name: str, description: str
+) -> None:
+    if not _project_name_is_available(client, project_name):
+        raise ProjectExists(f"Project with name: `{project_name}` already exists")
+
+    if project_name.endswith(" "):
+        raise ProjectValidationError(
+            f"Project name must not end with a tailing whitespace: `{project_name}`"
+        )
+
+    if not description:
+        raise ProjectValidationError("Project description is required")
+
+
+def _validate_tag_value(tag_value, tag_datatype) -> None:
+    if tag_datatype == "NUMBER":
+        try:
+            float(tag_value)
+        except ValueError:
+            raise InvalidMissionMetadata(f"Value '{tag_value}' is not a valid NUMBER")
+    elif tag_datatype == "BOOLEAN":
+        if tag_value.lower() not in {"true", "false"}:
+            raise InvalidMissionMetadata(
+                f"Value '{tag_value}' is not a valid BOOLEAN (expected 'true' or 'false')"
+            )
+    else:
+        pass  # any string is fine
+    # TODO: add check for LOCATION tag datatype
+
+
 def _get_metadata_type_id_by_name(
     client: AuthenticatedClient, tag_name: str
-) -> Optional[UUID]:
+) -> Tuple[Optional[UUID], str]:
     resp = client.get(TAG_TYPE_BY_NAME, params={"name": tag_name, "take": 1})
 
     if resp.status_code in (403, 404):
         return None
 
     resp.raise_for_status()
+    try:
+        data = resp.json()["data"][0]
+    except IndexError:
+        return None, None
 
-    data = resp.json()[0]
-    return UUID(data["uuid"], version=4)
+    return UUID(data["uuid"], version=4), data["datatype"]
 
 
 def _get_tags_map(
@@ -309,9 +481,10 @@ def _get_tags_map(
     # why are we using metadata type ids as keys???
     ret = {}
     for key, val in metadata.items():
-        metadata_type_id = _get_metadata_type_id_by_name(client, key)
+        metadata_type_id, tag_datatype = _get_metadata_type_id_by_name(client, key)
         if metadata_type_id is None:
             raise InvalidMissionMetadata(f"metadata field: {key} does not exist")
+        _validate_tag_value(val, tag_datatype)
         ret[metadata_type_id] = val
     return ret
 
